import os
import bpy, math
import mathutils                                                                                                                  
from mathutils import Vector, Matrix
from bpy.props import *
from xml.dom import minidom, Node
from xml.dom.minidom import parseString
#from priodict import priorityDictionary

#global variables
slovnik = {}
bones_names = []
pose_frame = []
frame_dist = 10
next_frame = 0
key_frames = []
incident_matrix = []
pose_AtKeyFrame = []
start_frame = 0
bignum = 100000 


def graphwalk(armature, treshold, start, end):
    global incident_matrix
    d = dict()
    size = len(incident_matrix)
    for i in range(0,size):
      d[i] = {}
      for j in range(0,size):
          help = dict()
          value = incident_matrix[i][j]
          if value<treshold:
              help[j] = value   
              d[i].update(help)
    print(d)
    if  sameComponent(d,start, end):
          print("Same component")
          result = shortestpath(d,start,end)
          distance = result[0]
          path = result[1]
          print("low cost path:")
          print(path)
          print("cost:")
          print(distance)
          animate(armature,path)
    else:
          print("Not in same component")



def shortestpath(graph,start,end,visited=[],distances={},predecessors={}):
    global bignum
    if start==end:
        path=[]
        while end != None:
            path.append(end)
            end=predecessors.get(end,None)
        return distances[start], path[::-1]
    if not visited: distances[start]=0
    for neighbor in graph[start]:
        if neighbor not in visited:
            neighbordist = distances.get(neighbor,bignum)
            tentativedist = distances[start] + graph[start][neighbor]
            if tentativedist < neighbordist:
                distances[neighbor] = tentativedist
                predecessors[neighbor]=start
    visited.append(start)
    unvisiteds = dict((k, distances.get(k,bignum)) for k in graph if k not in visited)
    closestnode = min(unvisiteds, key=unvisiteds.get)
    return shortestpath(graph,closestnode,end,visited,distances,predecessors)

def component_peak(graph, start, component=[]):
    component = component+[start]
    for node in graph[start]:
      if not node in component:
        component = component_peak(graph, node, component)
    return component

def sameComponent(graph,vertex1, vertex2):
    if vertex2 in component_peak(graph, vertex1):
        return True
    return False

def compare_pose(pose1_id, pose2_id, treshold):
    global bones_names
    global pose_AtKeyFrame
    value = 0
    pose1 = pose_AtKeyFrame[pose1_id]
    pose2 = pose_AtKeyFrame[pose2_id]
    for i in range(0,len(pose1)):
          matrix1 = pose1[i].to_3x3()
          matrix2 = pose2[i].to_3x3()
          if not compare_matrices(matrix1, matrix2, treshold):
             value = value +1
             print("different")
    return value      
  
def compare_matrices(matrix_first, matrix_second, tolerance):
  col = matrix_first.col_size
  row = matrix_first.row_size 
  if matrix_first.col_size != matrix_second.col_size or matrix_first.row_size != matrix_second.row_size:
    print('Dimensions of matrices is not equal!')
    return False
    
  for c in range(0,col):
    for r in range(0,row):
      if (abs(matrix_first[r][c] - matrix_second[r][c]) > tolerance):
        return False
  return True


def fill_incident_matrix(threshold):
    global incident_matrix
    global pose_AtKeyFrame
    size =  len(pose_AtKeyFrame)
    for i in range(0,size):
        for j in range(0,size):
             incident_matrix[i][j] = compare_pose(i, j, threshold)
             

def createStruct(rows,cols):
    global incident_matrix
    global pose_AtKeyFrame
    incident_matrix = [[1 for x in range(rows)] for x in range(rows)]
    print(incident_matrix)
    pose_AtKeyFrame = [[0 for i in range(cols)] for j in range(rows)]
    print ("structure")
    print (pose_AtKeyFrame)
        
def createStructure(source):
    global key_frames
    global pose_AtKeyframe
    global incident_matrix
    incident_matrix = []
    
    incident_matrix = [[1 for x in range(len(key_frames))] for x in range(len(key_frames))]
    print(incident_matrix)
    #array of lists which contain poses in keyframes from input
    for i in range(0,len(key_frames)):
        #pose_AtKeyFrame.append([])
        getpose_atFrame(source,key_frames[i])
        pose_AtKeyFrame.append(pose_frame)
        
def createArm(name, origin, boneTable):        # not used in final script, in further version used for debug
    # Create armature and object
    bpy.ops.object.add(
        type='ARMATURE', 
        enter_editmode=True,
        location=origin)
    ob = bpy.context.object
    ob.show_x_ray = True
    ob.name = name
    amt = ob.data
    amt.name = name+'Amt'
    amt.show_axes = True
 
    # Create bones
    bpy.ops.object.mode_set(mode='EDIT')
    for (bname, pname, vector) in boneTable:        
        bone = amt.edit_bones.new(bname)
        if pname:
            parent = amt.edit_bones[pname]
            bone.parent = parent
            bone.head = parent.tail
            bone.use_connect = False
            (trans, rot, scale) = parent.matrix.decompose()
        else:
            bone.head = (0,0,0)
            rot = Matrix.Translation((0,0,0))	# identity matrix
        bone.tail = rot * Vector(vector) + bone.head
    bpy.ops.object.mode_set(mode='OBJECT')
    return ob

def poseRig(ob, poseTable):                                       # not used in final script, in further version used for debug
    bpy.context.scene.objects.active = ob
    bpy.ops.object.mode_set(mode='POSE')
 
    for (bname, axis, angle) in poseTable:
        pbone = ob.pose.bones[bname]
        # Set rotation mode to Euler XYZ, easier to understand
        # than default quaternions
        pbone.rotation_mode = 'XYZ'
        # Documentation bug: Euler.rotate(angle,axis):
        # axis in ['x','y','z'] and not ['X','Y','Z']
        pbone.rotation_euler.rotate_axis(axis, math.radians(angle))
    bpy.ops.object.mode_set(mode='OBJECT')
 
def run(origo,name):                                                    # not used in final script, in further version used for debug
    origin = Vector(origo)
    # Table of bones in the form (bone, parent, vector)
    # The vector is given in local coordinates
    boneTable1 = [
        ('Base', None, (1,0,0)),
        ('Mid', 'Base', (1,0,0)),
        ('Tip', 'Mid', (0,0,1))
    ]
    bent = createArm(name, origin, boneTable1)
     
    # Pose rig
    poseTable1 = [
        ('Tip', 'Y', 45),
        ('Mid', 'Y', 45),
        ('Base', 'Y', 45)
    ]
    poseRig(bent, poseTable1)
 



def anim():                                                # not used in final script, in further version used for debug
    for bone in bpy.data.armatures[0].bones:
        bone.select = True
    bpy.ops.object.mode_set(mode='POSE')  
    # select the target object
    t = bpy.data.objects[0]
    # prepare a scene
    scn = bpy.context.scene
    scn.frame_start = 1
    scn.frame_end = 100
    
    # 1st keyframe
    scn.frame_set(1)
    t.pose.bones[2].rotation_mode = 'XYZ'
    t.pose.bones[2].rotation_euler.rotate_axis('X', math.radians(0))
    t.pose.bones[0].rotation_mode = 'XYZ'
    t.pose.bones[0].rotation_euler.rotate_axis('X', math.radians(0))
    bpy.ops.anim.keyframe_insert_menu(type='LocRot')
    
    # 2nd keyframe
    scn.frame_set(40)
    t.pose.bones[2].rotation_mode = 'XYZ'
    t.pose.bones[2].rotation_euler.rotate_axis('X', math.radians(45))
    t.pose.bones[0].rotation_mode = 'XYZ'
    t.pose.bones[0].rotation_euler.rotate_axis('X', math.radians(45))
    #t.pose.bones[1].rotation_quaternion = (1,0.5,0,0)
    bpy.ops.anim.keyframe_insert_menu(type='LocRot')
    
    # play animation   
    bpy.ops.screen.animation_play()
    
    #ANALYZE ANIMATION
    
    #set of bone name
    bones_names = []
    for bone in bpy.data.objects[0].pose.bones:
        bones_names.append(bone.name)
    print(bones_names)
    
    #set of bone location
    bones_location = []
    for bone in bpy.data.objects[0].pose.bones:
        bones_location.insert(bones_names.index(bone.name),bone.location)
    print(bones_location)
    
    #animation frame range
    frame_range = bpy.data.objects[0].animation_data.action.frame_range
    print(frame_range)
    begin = frame_range[0]
    end = frame_range[1]
    
    # find key frames
    scn.frame_set(begin - 1)
    global key_frames
    key_frames = []
    key = -1
    before = -2
    while key != before:
           before = key
           bpy.ops.screen.keyframe_jump()
           key = scn.frame_current
           if key != before:
              key_frames.append(key)       
    print(key_frames)
    bpy.ops.object.mode_set(mode='OBJECT')


def analyze_animation():                                    # analyze initial animation, extracting data
    #set of bone name
    global bones_names
    for bone in bpy.data.objects[0].pose.bones:
        bones_names.append(bone.name)
    print(bones_names)
    #animation frame range
    scn = bpy.context.scene
    frame_range = bpy.data.objects[0].animation_data.action.frame_range
    print(frame_range)
    begin = frame_range[0]
    end = frame_range[1]
    
    # find key frames
    scn.frame_set(begin - 1)
    global key_frames
    key_frames = []
    key = -1
    before = -2
    while key != before:
           before = key
           bpy.ops.screen.keyframe_jump()
           key = scn.frame_current
           if key != before:
              key_frames.append(key)       
    print(key_frames)
    bpy.ops.object.mode_set(mode='OBJECT')
    
def retarget(source, target,fromFrame,toFrame, scena):
    global pose_frame
    bpy.ops.object.mode_set(mode='POSE')
    size = 0
    for p in target.pose.bones:
        size += 1
    help1 = size
    while size>0:
        target.pose.bones[size-1].matrix = pose_frame[help1 - size]
        size -= 1

def retargeting(source, target, fromframe,toFrame, scena):
    global pose_frame
    bpy.ops.object.mode_set(mode='POSE')
    size = 0
    for bone in target.pose.bones:
        bone.matrix = pose_frame[size]
        size +=1
        
def retarget1(arm1, arm2, atFrame):
    bpy.context.scene.frame_set(atFrame)
    bpy.ops.object.mode_set(mode='POSE')
    index = 0
    size = 0
    for p in arm1.pose.bones:
        size += 1
    while size>0:
        poza = arm1.pose.bones[size-1].matrix.copy()
        arm2.pose.bones[size-1].matrix = poza
        size -= 1

def retarget2(arm1,arm2,atFrame): 
    bpy.context.scene.frame_set(atFrame)
    for bone in arm1.pose.bones:
        arm2.pose.bones[bone.name].matrix = bone.matrix.copy()
        
def pose_atFrame(arm,frame):                #compute pose at given frame
    global pose_frame
    pose_frame = []
    bpy.context.scene.frame_set(frame)
    bpy.ops.object.mode_set(mode='POSE')
    size = 0
    for p in arm.pose.bones:
        size += 1
    help = size
    while size>0:
        poza = arm.pose.bones[size-1].matrix.copy()
        pose_frame.append(poza)
        size -= 1         
def getpose_atFrame(arm, frame):
    global pose_frame
    pose_frame = []
    bpy.context.scene.frame_set(frame)
    bpy.ops.object.mode_set(mode='POSE')
    for p in arm.pose.bones:
        poza = p.matrix.copy()
        pose_frame.append(poza)
      

def animate2(source, target, key_frames):         # not used in final script, in further version used for debug
   
    global pose_frame
    global next_frame
    global frame_dist
    global start_frame
    bpy.ops.object.mode_set(mode='POSE')  
    for bone in bpy.data.armatures[1].bones:
        bone.select = True
    scn = bpy.context.scene
    scn.frame_start = 1
    scn.frame_end = 100
    before = 0
    for f in key_frames:
        getpose_atFrame(source,f)
        print('matrix')
        print(pose_frame)
        scn.frame_set(start_frame+next_frame)
        retargeting(source,target,f,f+50,scn)    
        bpy.ops.anim.keyframe_insert_menu(type='LocRot')
        next_frame = next_frame + frame_dist
        
    bpy.ops.screen.animation_play()

def ret_source(target,frame):                          #retarget pose
    global pose_frame
    bpy.ops.object.mode_set(mode='POSE')
    size = 0
    for bone in target.pose.bones:
        bone.matrix = pose_AtKeyFrame[frame][size]
        size +=1

def animate(target, key_frames):                         #create animation
   
    global pose_frame
    global next_frame
    global frame_dist
    global start_frame
    bpy.ops.object.mode_set(mode='POSE')  
    for bone in bpy.data.armatures[0].bones:
        bone.select = True
    scn = bpy.context.scene
    scn.frame_start = 1
    scn.frame_end = 100
    before = 0
    next_frame = 0
    for f in key_frames:
        print('matrix')
        print(pose_frame)
        scn.frame_set(start_frame+next_frame)
        ret_source(target,f)    
        bpy.ops.anim.keyframe_insert_menu(type='LocRot')
        next_frame = next_frame + frame_dist
        
    bpy.ops.screen.animation_play()


def matrixToString(matrix):
    help = ""
    prvok = 0
    for i in range(0,4):
        for j in range(0,4):
            prvok = prvok + 1
            help = help + str(matrix[i][j])
            if(prvok!=16):
                help = help + ', '
    return help  
          
def createXML():
    global key_frames
    global pose_AtKeyFrame
    doc = minidom.Document()
    
    doc.appendChild(doc.createComment("3D model bones"))
    
    frames = doc.createElement('frames')
    frames.setAttribute("count",str(len(key_frames)))
    frames.setAttribute("count_bones",str(len(bpy.data.objects[0].pose.bones)))
    
    doc.appendChild(frames)
    for j in range(0,len(key_frames)):
        name = "frame"
        fr = doc.createElement(name)
        fr.setAttribute("id_keyframe", str(j))
        frames.appendChild(fr)
        for i in range(0,len(pose_AtKeyFrame[j])):
             name = "bone"
             obj = doc.createElement(name)
             obj.setAttribute("id_bone", str(i))
             obj.setAttribute("id_keyframe", str(j))
             obj.appendChild(doc.createTextNode(matrixToString(pose_AtKeyFrame[j][i])))
             fr.appendChild(obj)
    
    print (doc.toprettyxml(indent = '   '))
    file = open("bones.xml", 'w')
    file.write(doc.toxml())

def parseXML(file_path):
    global pose_AtKeyFrame
       
    #open the xml file for reading:
    file = open( file_path,'r')
    #convert to string:
    data = file.read()
    #close file because we dont need it anymore:
    file.close()
    #parse the xml you got from the file
    dom = parseString(data)
    element_node_list = dom.getElementsByTagName("frames")
    element_n = element_node_list[0]
    count_keyframes = int(element_n.attributes["count"].value)
    count_bones = int(element_n.attributes["count_bones"].value)
    createStruct(count_keyframes,count_bones)
    element_node_list = dom.getElementsByTagName("bone") 
    for i in range(0,element_node_list.length):
         element_n = element_node_list[i]
         id_keyframe = int(element_n.attributes["id_keyframe"].value)
         id_kost = int(element_n.attributes["id_bone"].value)
         data = element_n.firstChild.data
         kf = []
         array = data.rsplit(',')
         for x in array:
            x = float(x)
            kf.append(x)
         newMat = Matrix(((kf[0], kf[1], kf[2], kf[3]),
                 (kf[4],kf[5],kf[6], kf[7]),
                 (kf[8], kf[9], kf[10], kf[11]) ,
                 (kf[12], kf[13], kf[14], kf[15])))
         pose_AtKeyFrame[id_keyframe][id_kost] = newMat 
         print("#---BONE---#")
         print (id_keyframe)
         print (id_kost)
         print (data)
         print (newMat) 
    print ("structure")
    print (pose_AtKeyFrame)
 
#
#    Store properties in the active scene
#
def initSceneProperties(scn):
    bpy.types.Scene.Start = IntProperty(
        name = "Starting Frame", 
        description = "Enter starting frame")
    scn['Start'] = 0
    
    bpy.types.Scene.End = IntProperty(
        name = "Ending Frame", 
        description = "Enter ending frame")
    scn['End'] = 0
    
    bpy.types.Scene.Dist = IntProperty(
        name = "Distance of key frames", 
        description = "Enter count of frames")
    scn['Dist'] = 10
 
    bpy.types.Scene.MyFloat = FloatProperty(
        name = "Threshold", 
        description = "Set threshold value",
        default = 10,
        min = 0,
        max = 100)
    
    bpy.types.Scene.MyFl = FloatProperty(
        name = "Tolerance", 
        description = "Set tolerance value",
        default = 0.5,
        min = 0,
        max = 2)
 
    bpy.types.Scene.MyBool = BoolProperty(
        name = "Graph walk", 
        description = "True or False?")
    scn['MyBool'] = False
 
    bpy.types.Scene.MyEnum = EnumProperty(
        items = [('Eine', 'type1', 'One'), 
                 ('Zwei', 'type2', 'Two'),
                 ('Drei', 'type3', 'Three')],
        name = "Evalueting method")
    scn['MyEnum'] = 2
 
    bpy.types.Scene.MyString = StringProperty(                         
        name = "sequence of frames"
       )
    scn['MyString'] =  str(scn['Start']) + "," +  str(scn['End'])
    return
 
initSceneProperties(bpy.context.scene)
 
#
#    Menu in UI region
#
class VIEW3D_PT_custompathmenupanel(bpy.types.Panel):
    bl_label = "Motion Graph panel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "TOOLS"
 
    def draw(self, context):
        layout = self.layout
        scn = context.scene
        layout.prop(scn, 'MyPath')
        layout.prop(scn, 'Start')
        layout.prop(scn, 'End')
        layout.prop(scn, 'Dist')
        layout.prop(scn, 'MyFloat')
        layout.prop(scn, 'MyFl')
        layout.prop(scn, 'MyBool')
        layout.prop(scn, 'MyEnum')
        layout.prop(scn, 'MyString')
        layout.operator("input.animation")
        layout.operator("import.file")
        layout.operator("create.anim")
 
#
#    The button prints the values of the properites in the console.
#
 
class OBJECT_OT_PrintPropsButton(bpy.types.Operator):
    bl_idname = "import.file"
    bl_label = "Import Motion Graph Structure from XML"
 
    def execute(self, context):
        global frame_dist
        global incident_matrix
        global pose_AtKeyFrame
        scn = context.scene
        printProp("Int:    ", 'MyInt', scn)
        printProp("Float:  ", 'MyFloat', scn)
        printProp("Bool:   ", 'MyBool', scn)
        printProp("Enum:   ", 'MyEnum', scn)
        printProp("String: ", 'MyString', scn)
        frame_dist = scn['Dist']
        start_frame = scn['Start']
      
        name_of_selected_object = bpy.context.active_object.name #choose selected object
        armature = bpy.data.objects[name_of_selected_object]

        bpy.context.scene.objects.active = armature
        bpy.ops.object.mode_set(mode='POSE')
        
        parseXML(bpy.context.scene.MyPath)
        fill_incident_matrix(scn['MyFl'])
        frames = []
        for i in range(0,len(pose_AtKeyFrame)):
            frames.append(i)
        animate(armature,frames)
        
       
        return{'FINISHED'}    

class OBJECT_OT_Input_Button(bpy.types.Operator):
    bl_idname = "input.animation"
    bl_label = "Export animation to XML"
   
 
    def execute(self, context):
      global key_frames
      key_frames = []
      scn = context.scene
      bpy.ops.wm.console_toggle()
      bpy.ops.object.mode_set(mode='OBJECT')
      print ("Globalna premenna key frames: ") 
      print (key_frames)
      analyze_animation()
      createStructure(bpy.data.objects[0])
      createXML()
      return{'FINISHED'}

class OBJECT_OT_Input_Button(bpy.types.Operator):
    bl_idname = "create.anim"
    bl_label = "Create new animation"
   
 
    def execute(self, context):
        global frame_dist
        global incident_matrix
        scn = context.scene
        
        frame_dist = scn['Dist']
        start_frame = scn['Start']
        x = parseStr(scn["MyString"])
        print(x)
        frames = []
        frames = parseStr(scn["MyString"])
        name_of_selected_object = bpy.context.active_object.name #vyberieme objekt, ktory je oznaceny
        armature = bpy.data.objects[name_of_selected_object]

        bpy.context.scene.objects.active = armature
        bpy.ops.object.mode_set(mode='POSE')
        armature.animation_data_clear()
        if scn['MyBool']:
            fill_incident_matrix(scn['MyFl'])      
            graphwalk(armature,scn['MyFloat'], scn['Start'], scn['End'])
            print("graphwalk")
            print(scn['Start'])
            print(scn['End'])
        else:
            animate(armature,frames)
            
        return{'FINISHED'}    


bpy.types.Scene.MyPath = StringProperty(name="XML file path: ",
    attr="custompath",# this a variable that will set or get from the scene
    description="simple file path",
    maxlen= 1024,
    subtype='FILE_PATH',
    default= "")#this set the text



    
  
def printProp(label, key, scn):
    try:
        val = scn[key]
    except:
        val = 'Undefined'
    print("%s %s" % (key, val))
 
def parseStr(input):                #parsing of string
    kf = []
    array = input.rsplit(',')
    for x in array:
        x = int(x)
        kf.append(x)
    return kf 

#    Registration
bpy.utils.register_module(__name__)